#! /usr/bin/perl

#
# ---------------------------------------------------------------------
#
# dock: Script for Docking Studies (c) 2005, 2006 - Csar L. Araujo
#
# ---------------------------------------------------------------------
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software 
# Foundation, Inc., 51 Franklin Street, Fifth Floor, 
# Boston MA 02110-1301 USA
#
# ---------------------------------------------------------------------
#
# Based on tutorial:
# Ref.: Yu, H.; Rosen, M.K.; Saccomano, N. A.; Phillips, D.; Volkmann, R. A.; 
# 	Schreiber, S. L.;: Sequential assignment and structure determination
#	of spider toxin omega-Aga-IVB. Biochemistry 32 pp. 13123 (1993)
#
# http://www2.umdnj.edu/~kerrigje/pdf_files/fwspidr_tutor.pdf
#
# ---------------------------------------------------------------------
#
# Contact to the author: 
# 
# Lic. Csar L.Araujo
# e-mail: cesar.araujo@oulu.fi
# Research Center for Molecular Endocrinology
# WHO CCR / Biocenter Oulu / University of Oulu
# and Biocomputing Group / Dept. of Biochemistry / University of Oulu
#
# ---------------------------------------------------------------------
#

use strict;
use Getopt::Long;
use File::Spec;

my %Option = ();


# Declare global file name variables

my $MPI_NP_PARAM;	# Number of processor used in the cluster
my $MPI_NT_PARAM;	# Number of threads instantiated by mdrun
my $MDRUN;		# GROMACS mdrun program according to the cluster used
my $MPI_PREFIX;		# Prefix for run "mdrun" program (e.g. "mpirun -np 4")
my $NOHUP;		# You can use it to select ("nohup") or deselect ("") nohup use
my $BACKGROUND;		# You can use it to select ("&") or deselect ("") runs in background mode
my $GROMPP_EXTRA;	# Used for extra options on grompp (e.g. -sort and -shuffle)
my $GROMACS_FF;		# GROMACS Force Field for pdb2gmx tool

my $HOSTS_FILE_DIR;	# Relative directory path for lamboot hosts definition file
my $PROJECT_DIRECTORY;	# General project directory
my $PATH_PREFIX; 	# The results will be saved in different directories   
my $MD_MDP_FILE;	# Parameters file for molecular dynamics simulation
my $PR_MDP_FILE;        # Parameters file for position restraints
my $GENION_LOG_FILE;	# Log file for genion program (used in -ionsolvated mode)
my $PDB_FILE;		# PDB file of protein
my $PDB2GRO_EXTRA;	# Extra options for pdb2gmx command, e.g, -his -asp, etc.

my $EM_MDP_FILE;	# Parameters file for energy minimization
my $EM_TPR_FILE;	
my $EM_TRR_FILE;
my $EM_LOG_FILE;
my $EM_EDR_FILE;
    
my $PR_TPR_FILE;
my $PR_TRR_FILE;
my $PR_EDR_FILE;
my $PR_LOG_FILE;

my $MD_TPR_FILE;
my $MD_TRR_FILE;
my $MD_EDR_FILE;
my $MD_LOG_FILE;	# Log file for simulation step

my $PDB2GRO_GRO_FILE;	# Coordinates file from PDB
my $GENBOX_GRO_FILE;   	# Coordinates file from genbox program (used in -solvated and -ionsolvated modes)
my $EM_GRO_FILE;
my $PR_GRO_FILE;
my $MD_GRO_FILE;
my $RESULT_GRO_FILE;

my $COMPLEX_GRO_FILE;

my $TOPOLOGY_FILE;	# Topology file

my $LIGAND_ITP_FILE;	# ITP file for ligand molecule
my $LIGAND_ITP_INCLUDE;	# Include path for topology file
my $LIGAND_GRO_FILE;	# DRGFIN.GRO file generated by PRODRG server
my $LIGAND_NAME;	# Name given by PRODRG to the ligand (by default is "DRG")
my @LIGAND_GROWING_TOPOLOGIES = ();

my $RESTART_FILE;		# TPR filename for restart after crash
my $RESTART_DEFAULT_NAME;	# Default filenames for mdrun after restart

my $FILE_PREFIX; 

my $DRG_ATOM_COUNT;	# Number of atoms in the ligand
my $PROT_ATOM_COUNT;

my $BOX_DISTANCE = 0.5;	# Distance to the periodic box's edge

my $LUCK;		# Used to test the successful of configuration file

# Global variables
my $element = "";
my $opt;
my $val;
my @lines;
my @coords = ();
my $i = 0;
my $j;
my $k;
my $pcount;
my $lastatomindex;
my $lastresindex;
my $aux;
my $HostName;
my $ngstep;
my $prev_ligand_topol;

# Parsing arguments

GetOptions(
	    "config=s"=>\$Option{"config"},
	    "mode=s"=>\$Option{"mode"},
	    "dprefix=s"=>\$Option{"dprefix"},
	    "iontype=s"=>\$Option{"iontype"},
	    "ionname=s"=>\$Option{"ionname"},
	    "continue=s"=>\$Option{"continue"},
	    "ioncount=i"=>\$Option{"ioncount"},
	    "docking"=>\$Option{"docking"},
	    "complex"=>\$Option{"complex"},
	    "continue_from_crash"=>\$Option{"continue_from_crash"},
	    "extend=i"=>\$Option{"extend"},
	    "until=i"=>\$Option{"until"},
	    "np=i"=>\$Option{"np"},
	    "x"=>\$Option{"x"},
	    "growing_ligand"=>\$Option{"growing_ligand"}
);

if ($Option{"config"} eq "") {
    die "Please, specify a config filename.";
}

# We show the script's arguments

print_log("\n################################################\n");
print_log("Starting process on " . localtime(time)."\n");
print_log("\n################################################\n");

print_log("Running script with the following options:\n");
while (($opt,$val)=each(%Option)) {
    print_log("$opt = $val\n");
}

if ($Option{"mode"} eq "invacuo") {
    print_log("I'm going to prepare stuffs for IN VACUO simulation...\n");
}

if ($Option{"mode"} eq "solvated") {
    print_log("I'm going to prepare stuffs for SOLVATED simulation...\n");
}

if ($Option{"mode"} eq "ionsolvated") {
    print_log("I'm going to prepare stuffs for ION SOLVATED simulation...\n");
}

# MPI options

# First check the file with nodes definition

$aux = GetNP();

if ($aux > 0) {
    # If the file exists, override the np option
    print_log("Running on NP = $aux nodes\n");
    $Option{"np"} = $aux;
} else {
    print_log("Warning: there isn't any node definition file. Are you running outside the cluster?\n");
}

if ($Option{"np"} > 0) {
    $MPI_NP_PARAM = "-np " . $Option{"np"};
    $GROMPP_EXTRA = ""; #"-sort -shuffle";
    $MPI_PREFIX = "mpirun $MPI_NP_PARAM C ";
    print_log("MPI_PREFIX = $MPI_PREFIX\n");
}

if ($Option{"nt"} > 0) {
    $MPI_NT_PARAM = "-nt " . $Option{"nt"};
}

# Particular file names

$LIGAND_NAME = "DRG"; # We assign the default value
$MDRUN = "mdrun";
$NOHUP = "nohup";
$BACKGROUND = "&";

if (ReadConfigFile($Option{"config"})) {
    print_log("Configuration file successfully read.\n");
} else {	
    die "Error reading configuration file !!!";
}

if ($NOHUP ne "") {
    $NOHUP .= " ";
}


#
# Batman begins :)
#

if ($Option{"continue_from_crash"} || $Option{"extend"} > 0 || $Option{"until"} > 0) {
    $aux = GetLastRestartIndex();
    my $PREV_TPR_FILE;
    my $PREV_TRR_FILE;
    my $PREV_EDR_FILE;
    my $NEW_RESTART_TPR_FILE;
    my $restart_time;
    my $extend_op;
    my $restart_op;
    my $until_op;
    
    if ($aux == 0) {
	$PREV_TPR_FILE = $MD_TPR_FILE;
	$PREV_TRR_FILE = $MD_TRR_FILE;
	$PREV_EDR_FILE = $MD_EDR_FILE;	
    } else {		
	$PREV_TPR_FILE = "$PATH_PREFIX/" . GetFileName($MD_TPR_FILE) . "$aux.tpr";
	$PREV_TRR_FILE = "$PATH_PREFIX/" . GetFileName($MD_TRR_FILE) . "$aux.trr";
	$PREV_EDR_FILE = "$PATH_PREFIX/" . GetFileName($MD_EDR_FILE) . "$aux.edr";		
    } 
    
    if ($Option{"extend"} > 0) {
	$extend_op = "-extend " . $Option{"extend"};
	print_log("Extending MD for " . $Option{"extend"} . " ps\n");
    }
    
    if ($Option{"until"} > 0) {
	$until_op = "-until " . $Option{"until"};
	print_log("Extending MD until time " . $Option{"until"} . " ps\n");
    }
    
    if ($extend_op eq "" &&  $until_op eq "") {
	# We are continuing from a crash, so we avoid the last frame
        $restart_time = GetRestartTime($PREV_TRR_FILE);
	print_log("Trying to continue after a crash !!!\n");
        print_log("Continuing from time $restart_time ps\n");
	$restart_op = "-time $restart_time";
    }
    
    $NEW_RESTART_TPR_FILE = "$PATH_PREFIX/" . GetFileName($MD_TPR_FILE) .($aux+1). ".tpr";
    
    Execute("tpbconv $restart_op $extend_op $until_op -s $PREV_TPR_FILE -f $PREV_TRR_FILE -e $PREV_EDR_FILE -o $NEW_RESTART_TPR_FILE");
    Execute($NOHUP.$MPI_PREFIX."$MDRUN -s $NEW_RESTART_TPR_FILE -deffnm $PATH_PREFIX/" . GetFileName($MD_TRR_FILE) . ($aux+1)." ".$BACKGROUND);
    goto EXIT_PROGRAM;
}

if (!($Option{"continue"} eq "")) {
    goto CONTINUE;
}

# Create the coordinates and topology file

if (-e $PDB_FILE) {
    print_log("I found a protein PDB file so I'm going to convert it to GROMACS coordinate file format.\n");
    if (chdir($PATH_PREFIX)) {
	#Execute("pdb2gmx -ignh $PDB2GRO_EXTRA -ff $GROMACS_FF -f $PDB_FILE -o $PDB2GRO_GRO_FILE -p $TOPOLOGY_FILE -i $PATH_PREFIX/posre.itp");
	Execute("pdb2gmx -ignh $PDB2GRO_EXTRA -ff $GROMACS_FF -f ../$PDB_FILE -o ../$PDB2GRO_GRO_FILE -p ../$TOPOLOGY_FILE -i posre.itp");
	chdir("..");	
    } else {
        die("Cannot change to $PATH_PREFIX subdirectory!!!\n");	
    }
}

if ($Option{"docking"}) {

    if (!$Option{"complex"}) {
        # If we are performing docking we need to make some dirty job with files.
	
	print_log("Performing docking preprocessing...\n");
    
        # Add the ligand coordinates to the new GRO file

	print_log("Adding ligand $LIGAND_GRO_FILE to $PDB2GRO_GRO_FILE\n");
        print_log("Opening $LIGAND_GRO_FILE\n");
        print_log("Opening $PDB2GRO_GRO_FILE\n");
	print_log("Opening $PDB2GRO_GRO_FILE.bak\n");
        open(ARCH,$LIGAND_GRO_FILE);
	open(ARCHPROT,$PDB2GRO_GRO_FILE);
        open(AOUT,">$PDB2GRO_GRO_FILE.bak");

        # The first line contains the title so we just write it to the output file.
	$_ = <ARCHPROT>;
        print AOUT $_;

	# The second line contains the atoms count.
        chop($PROT_ATOM_COUNT = <ARCHPROT>);

	# It's the same for the ligand file
        $j = <ARCH>;
	chop($DRG_ATOM_COUNT = <ARCH>);


        # We write the sum of atoms 'cause we are adding the ligand
	print AOUT $PROT_ATOM_COUNT + $DRG_ATOM_COUNT;
        print AOUT "\n";
	print_log("Total atom count:" . $PROT_ATOM_COUNT . "+" . $DRG_ATOM_COUNT . "\n");

        $pcount = 0;
	$i = <ARCHPROT>;
        $j = <ARCHPROT>;
	print AOUT $i;
        print AOUT $j;
	$pcount += 2;
        while($k = <ARCHPROT>) {        	
		$pcount++;
	
    	    if (!eof(ARCHPROT) && ($pcount-2) < ($PROT_ATOM_COUNT - 2)) {
		print AOUT $k;	    
		    #print $i;
    	    } else {
		print_log("Reaching end of file...\n");
	
		    # We are adding the ligand at the end of the file, so we need to renumbering it in order to provide consecutive numbers
		# for atoms and residue.
    	        $k =~ /\s+([0-9]+)\S+\s+\S+\s+([0-9]+).+/;
		$lastresindex = $1 + 1;
		$lastatomindex  = $2;
	
		print AOUT $k;
	    
		my $lac = 0; # ligand atom count
		while($k = <ARCH>) {
		    $lastatomindex++;	    	    
		    if ($lac < $DRG_ATOM_COUNT) {
			$lac++;
    			$aux = (" " x (5-length($lastresindex))) . $lastresindex;
			$k =~ s/\s+[0-9]+/$aux/;
			$aux = (" " x (5-length($lastatomindex))) . $lastatomindex;
	    
			# The residue's name can be 5 chars long so we need 2 extra space chars due to "DRG" name given by PRODRG server.
			# $k =~ s/(\s+[0-9]+)(\S+)(\s+)(\S+)(\s+)([0-9]+)(.+)/\1\2  \3\4$aux\7/;
			$k =~ s/(\s*[0-9]+)(\S+)(\s*)(\S+)(\s*)([0-9]+)(.+)/\1\2\3\4$aux\7/;
			print AOUT $k;	    
			#print $k;
		    }
	    
		}
		print AOUT $k;
	    }
    
	}


	close(AOUT);
	close(ARCHPROT);
	close(ARCH);


	# Ok, I have added the ligand, so we rename the file to replace the original one.

	if (system("cp $PDB2GRO_GRO_FILE.bak $PDB2GRO_GRO_FILE")) {
	    die "Cannot rename $PDB2GRO_GRO_FILE.bak to $PDB2GRO_GRO_FILE in ".$Option{"mode"} ." mode !!!";
	}
	
    } else {
	if (system("cp $COMPLEX_GRO_FILE $PDB2GRO_GRO_FILE")) {
	    die "Cannot rename $COMPLEX_GRO_FILE to $PDB2GRO_GRO_FILE in ".$Option{"mode"} ." mode !!!";
	}
    }

    # Now we add the ligand to topology file

    AddSpecie(1,$LIGAND_NAME,$LIGAND_ITP_INCLUDE);
}


# Use truncated octahedral periodic box (0.5 nm)
Execute("editconf -bt octahedron -f $PDB2GRO_GRO_FILE -o $PDB2GRO_GRO_FILE -c -d $BOX_DISTANCE","Creating periodic box...");

#
# Retrieve the last line of out.gro to calculate the center of coordinates
#
open(ARCH,"$PDB2GRO_GRO_FILE");

@lines = <ARCH>;

 if ($lines[$#lines] =~ /([0-9\.]+)\s+([0-9\.]+)\s+([0-9\.]+).*/) {
    $coords[0] = $1;
    $coords[1] = $2;
    $coords[2] = $3;
    print "Extracting out.gro coordinates for center calculation: OK\n";
 } else {
    unless ($Option{"-x"}) {
	print "I CANNOT PARSE THE LAST LINE OF OUT.GRO!!! Perhaps the file is missing.\n";
    } else {
	die "I CANNOT PARSE THE LAST LINE OF OUT.GRO!!! Perhaps the file is missing.\n";
    }
 }

close(ARCH);

# Center the molecule   
Execute("editconf -f $PDB2GRO_GRO_FILE -o $PDB2GRO_GRO_FILE -center " . 
       eval($coords[0]/2.0) . " " . eval($coords[1]/2.0) . " " . eval($coords[2]/2.0),"Centering molecule in the box...");

CONTINUE:

if ($Option{"mode"} eq "solvated" || $Option{"mode"} eq "ionsolvated") {
    #
    # Processing for solvated system
    #
    
    if ($Option{"continue"} eq "RUN_POSITION_RESTRAINED") {
	goto RUN_POSITION_RESTRAINED;
    }
    
    if ($Option{"continue"} eq "RUN_DYNAMICS") {
	goto RUN_DYNAMICS;
    }

RUN_ENERGY_MINIMIZATION:        
    # Generate the solvated periodic box
    Execute("genbox -cp $PDB2GRO_GRO_FILE -cs spc216.gro -o $GENBOX_GRO_FILE -p $TOPOLOGY_FILE","Solvating the box...");

    $ngstep = 0;
    $prev_ligand_topol = $LIGAND_ITP_FILE;
RUN_NEXT_STEP:

    if ($Option{"growing_ligand"}) {
	print_log("Growing ligand step #$ngstep\n");
	ModifyLigandTopology($prev_ligand_topol, $LIGAND_GROWING_TOPOLOGIES[$ngstep]);
	$prev_ligand_topol = $LIGAND_GROWING_TOPOLOGIES[$ngstep];
    }    

    # Preprocessing  for energy minimization (em)
    Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -maxwarn 1000 -f $EM_MDP_FILE -c $GENBOX_GRO_FILE -p $TOPOLOGY_FILE -o $EM_TPR_FILE", "Preprocessing for energy minimization...");    
    
    if ($Option{"mode"} eq "ionsolvated") {
	#
	# Our model has a net charge. Therefore, we want to add counter ions.
	#
	
	Execute("genion -s $EM_TPR_FILE -o $EM_GRO_FILE -".$Option{"iontype"}."name ".$Option{"ionname"}." -n".
		$Option{"iontype"}." ".$Option{"ioncount"}." -g $GENION_LOG_FILE", "Adding counterions to our ion...");
	
	# Substract nncount from the total number of SOL molecules in the topology file
	
	print_log("Updating topology file...\n");
	AddCounterIon($Option{"ioncount"}, $Option{"ionname"});
	
    }
    
    if ($Option{"mode"} eq "ionsolvated") {    
        # We need to preprocess again for energy minimization (em) if -ionsolvated option is selected.
	# The reason is that $EM_TPR_FILE must be generated again in order to match the modified fws.top.
	# Otherwise, mdrun fails due to a mismatch in the coordinates count between atoms in topology file
	# and the previous $EM_TPR_FILE file. 
	Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -f $EM_MDP_FILE -c $EM_GRO_FILE -p $TOPOLOGY_FILE -o $EM_TPR_FILE", "Preprocessing for energy minimization again after adding ions...");
    }
        
    # Run energy minimization
    Execute($MPI_PREFIX . "$MDRUN $MPI_NP_PARAM $MPI_NT_PARAM -s $EM_TPR_FILE -o $EM_TRR_FILE -c $PR_GRO_FILE -g $EM_LOG_FILE -e $EM_EDR_FILE", "Running energy minimization...");

    #
    # This step is needed to equilibrate the solvent molecules
    #    

RUN_POSITION_RESTRAINED:
    # Preprocessing for Position-restrained molecular dynamics    
    Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -maxwarn 1000 -f $PR_MDP_FILE -c $PR_GRO_FILE -r $PR_GRO_FILE -p $TOPOLOGY_FILE -o $PR_TPR_FILE", "Preprocessing for position restraints...");

    # Run position-restrained molecular dynamics (pr)
    Execute($MPI_PREFIX . "$MDRUN $MPI_NP_PARAM $MPI_NT_PARAM -s $PR_TPR_FILE -o $PR_TRR_FILE -c $MD_GRO_FILE -g $PR_LOG_FILE -e $PR_EDR_FILE", "Running position restraints...");
    
    if ($Option{"growing_ligand"} && $ngstep < $#LIGAND_GROWING_TOPOLOGIES) {
	ExtractLastPRFrame();
	$ngstep++;
	goto RUN_NEXT_STEP;
    }    

RUN_DYNAMICS:    
    # Preprocessing for solvated model
    Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -f $MD_MDP_FILE -c $MD_GRO_FILE -r $MD_GRO_FILE -p $TOPOLOGY_FILE -o $MD_TPR_FILE", "Preprocessing for simulation...");

    # Run molecular dynamics for solvated model
    Execute($NOHUP . $MPI_PREFIX . "$MDRUN $MPI_NP_PARAM $MPI_NT_PARAM -s $MD_TPR_FILE -o $MD_TRR_FILE -c $RESULT_GRO_FILE -g $MD_LOG_FILE -e $MD_EDR_FILE ".$BACKGROUND, "Running simulation...");
        

} elsif ($Option{"mode"} eq "invacuo") {
    #
    # Processing for in vacuo system
    #
    
    # Process topology file for energy minimization (em)
    Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -f $EM_MDP_FILE -c $PDB2GRO_GRO_FILE -p $TOPOLOGY_FILE -o $EM_TPR_FILE", "Preprocessing for energy minimization...");
    
    # Run energy minimization
    Execute($MPI_PREFIX . "$MDRUN $MPI_NP_PARAM $MPI_NT_PARAM -s $EM_TPR_FILE -o $EM_TRR_FILE -c $PDB2GRO_GRO_FILE -g $EM_LOG_FILE -e $EM_EDR_FILE", "Running energy minimization...");

    # Preprocessing for in vacuo model
    Execute("grompp $MPI_NP_PARAM $GROMPP_EXTRA -f $MD_MDP_FILE -c $PDB2GRO_GRO_FILE -p $TOPOLOGY_FILE -o $EM_TPR_FILE", "Preprocessing for simulation...");
    
    # Run molecular dynamics for in vacuo model
    Execute($NOHUP . $MPI_PREFIX . "$MDRUN $MPI_NP_PARAM $MPI_NT_PARAM -s $EM_TPR_FILE -o $MD_TRR_FILE -c $PDB2GRO_GRO_FILE -g $MD_LOG_FILE -e $MD_EDR_FILE ".$BACKGROUND, "Running simulation...");

} else {
    print_log("Nothing else to do !!!\n");
}

if ($Option{"x"}) {
    print_log("Uffff... That was quite a lot work... Have a nice day !!! :)\n\n");
} else {
    print_log("That's all I can do for you...\n\n");
}

EXIT_PROGRAM:
exit(0);

#----------------------------------------------------------------------------
sub AddSpecie
{
    my $specie_count = @_[0];
    my $name = @_[1];
    my $include_file_name = @_[2];
    
    open(ARCH,">>$TOPOLOGY_FILE");
    print ARCH "$name\t\t$specie_count\n";
    close(ARCH);
    
    open(ARCH,$TOPOLOGY_FILE);
    open(AOUT,">$TOPOLOGY_FILE.bak");
    
    while(<ARCH>) {
	print AOUT $_;
        #if ($include_file_name <> "") {
    	    if ($_ =~ /Include forcefield parameters/) {
		    # It seems that GROMACS is reluctant for automation. Before include our ITP file is needed to include the forcefield specific ITP file.
		    $_ = <ARCH>;
		    print AOUT $_;
    		    print AOUT "#include \"$include_file_name\"\n";
		    print_log("Including $include_file_name\n");
    	    }
	#}
    }
    close(ARCHOUT);
    close(ARCH);
    
    if (system("cp $TOPOLOGY_FILE.bak $TOPOLOGY_FILE")) {
	#die "Cannot rename  $TOPOLOGY_FILE.bak to $TOPOLOGY_FILE in ionsolvated mode !!!";
    }
}

# Substract nncount from the total number of SOL molecules in the topology file
sub AddCounterIon
{
    my $added_count;
    my $name;
    my $success = 0;
    
    $added_count = @_[0];
    $name = @_[1];
    
    print_log("Modifying topology file $TOPOLOGY_FILE...\n");
    print_log("Trying to add $added_count molecules of $name\n");
    open(ARCH,$TOPOLOGY_FILE);
    open(AOUT,">$TOPOLOGY_FILE.bak");
    while(<ARCH>) {
        if ($_ =~ /^SOL(\s+)([0-9]+)/) {
	    $aux = eval($2 - $added_count);
	    print AOUT "SOL$1" . $aux  . "\n";
    	    print AOUT $name."$1$added_count"."\n";
	    print_log("$added_count of $name successfuly added !!!\n");
	    $success = 1;
	} else {
	    print AOUT $_;
	    
	}
    }
    
    close(AOUT);
    close(ARCH);
	
    if (system("cp $TOPOLOGY_FILE.bak $TOPOLOGY_FILE")) {
	die "Cannot rename  $TOPOLOGY_FILE.bak to $TOPOLOGY_FILE in ionsolvated mode !!!";
    }
    
    if (not $success) {
	if ($Option{"x"}) {
	    print_log("Warning: I couldn't add counterions !!!. Perhaps due to a failure in solvation step.\n");
	}
    }

}

#
# This function is used to execute (-x) or just print the GROMACS functions
#
sub Execute
{
    print_log("\n".@_[1]."\n");
    print_log(@_[0]."\n");
    
    if ($Option{"x"} == 1) {
	print(@_[0]."\n");
	if (system @_[0]) {
	    die "System error: $?\n";
	}
    } else {
	print(@_[0]."\n");
    }
    
}

sub print_log
{
    open(LOG_ARCH,">>dock.log");
    print @_;
    print LOG_ARCH @_;
    close(LOG_ARCH);
}

sub ReadConfigFile
{
    my $fname;
    my $line;
    my $str;
    
    open(CONFIG,@_[0]);
    
    if (eof(CONFIG)) {
	$_ = 0;
    }
    
    print_log("Reading configuration files...\n");
    
    while($line = <CONFIG>) {
        $str .= $line;
    }
        
    eval($str);
    
    if (CheckHealth()) {
        print_log("$LUCK\n");
    } else {
	print_log("Please, check configuration file.\n");
    }
    
    close(CONFIG);
    
    $_ = 1;
}

sub CheckHealth
{
    my $retval = 1;
    
    if (not -e $Option{"config"}) {
	print_log("Fatal error: project configuration file (".$Option{"config"}.") doesn't exist.\n");
	die;
    }
    
    if (not -d $PROJECT_DIRECTORY) {
	print_log("Fatal error: project directory ($PROJECT_DIRECTORY) doesn't exist.\n");
	die;
    }
    
    if (not -d "$PROJECT_DIRECTORY/data") {
	print_log("Fatal error: data directory ($PROJECT_DIRECTORY/data) doesn't exist.\n");
	die;
    }
    
    
    # If there are log files it stops to avoid undesired overridings

    if ($Option{"continue_from_crash"} == 0 && $Option{"continue"} eq "") {    
        TestLogFilesExistenceFor("mode");
    }
    
    if (not -e $PDB_FILE) {
	print_log("Warning: protein PDB file ($PDB_FILE) doesn't exist. Are you going to work just with the ligand?\n");
	if (not $PDB_FILE eq "") {
	    print_log("Fatal error: sorry \$PDB_FILE = \"$PDB_FILE\" is not empty and the file doesn't exist. If you are just working with the ligand, please just leave this variable empty.\n");
	    die;
	}
    }
    
    if (not -e $MD_MDP_FILE) {
	print_log("Fatal error: .MDP file for simulation ($MD_MDP_FILE) doesn't exist.\n");
	die;
    }
        
    if ($Option{"mode"} eq "solvated" or $Option{"mode"} eq "ionsolvated") {
	if (not -e $EM_MDP_FILE) {
	    print_log("Fatal error: .MDP file for energy minimization ($EM_MDP_FILE) doesn't exist.\n");
	    die;
	}
	
	if (not -e $PR_MDP_FILE) {
	    print_log("Fatal error: .MDP file for position restraints ($PR_MDP_FILE) doesn't exist.\n");
	    die;
	}
    }
    
    if (not -e $LIGAND_GRO_FILE) {
	print_log("Warning: ligand .GRO file ($LIGAND_GRO_FILE) doesn't exist. Are you going to work just with the protein?\n");
    }
    
    my $file_itp;
    
    if (not -e $LIGAND_ITP_FILE) {
	print_log("Warning: ligand .ITP file ($LIGAND_ITP_FILE) doesn't exist. Are you going to work just with the protein?\n");
    }
    
    if ($retval == 0) {
	die;
    }
    $_ = $retval;
}

sub GetNP
{
    my $HostName = $ENV{"HOSTNAME"};
    my @lines;
    my $nCpu = 0;

    if (-e $HOSTS_FILE_DIR."$HostName") {
	print_log("MPI nodes definition file found ($HostName)\n");
	open(ARCHIN,$HOSTS_FILE_DIR."$HostName");
	@lines = <ARCHIN>;
	for ($i = 0; $i <= $#lines; $i++) {
	    if ($lines[$i] =~ /.+\s+cpu=(.+)/) {
		$nCpu += $1;
	    } elsif ($lines[$i] =~ /.+/) {
		$nCpu++;
	    }
	    
	}
	print_log("Running on nodes:\n");
	print_log(@lines);
	print_log("Total CPU: $nCpu\n");
	print_log("\n");
        close(ARCHIN);
	
	return $nCpu;
    }
    
    return 0;
}

sub TestLogFilesExistenceFor
{
    my @log_files;
    
    print_log("Checking for the existence of previous log files\n");
    @log_files = glob($PROJECT_DIRECTORY."/".@_[0]."/*.log");
    
    print_log("Count of log files in ".@_[0]." directory: ".($#log_files+1)."\n");
    if ($Option{@_[0]} && ($#log_files+1) > 0) {
        print_log("Fatal error: the directory contains log files. Please delete the log files in order to continue.\n");
        die;
    }
}

sub GetLastRestartIndex
{
    my $maxidx;
    my $i;
    my @prev_files;
    
    @prev_files = glob($PROJECT_DIRECTORY."/".$Option{"mode"}."/$MD_TRR_FILE*.trr");
    
    $maxidx = 0;
    
    for ($i = 0; $i < $#prev_files+1; $i++) {
	if ($prev_files[$i] =~ /restart([0-9]+)\..*/) {
	    if ($1 > $maxidx) {
		$maxidx = $1;
	    }
	}
    }
    
    return $maxidx;    

}



sub GetRestartTime
{
    my $trr_file;
    my $lf;
    my $t;
    my $time_step;
    
    $trr_file = @_[0];
    
    $t = 0;
    
    if (system("gmxcheck -f $trr_file 2>$PATH_PREFIX/foo.1")) {
	die("GetRestartTime error!!! Reason: gmxcheck on \"$trr_file\" file failed.\n");
    }
    if (system("cat $PATH_PREFIX/foo.1 |grep \"Step\" >$PATH_PREFIX/foo.2")) {
	die("GetRestartTime error!!! Reason: cannot read gmxcheck output.\n");
    }
    
    open(ARCH, "$PATH_PREFIX/foo.1");
    
    while ($lf = <ARCH>) {
    
        if ($lf =~ /Last frame\s+([0-9]+) time\s+([0-9]+).*/) {
		$t = $2;
		print_log("Last frame found for $t ps\n");
		last;
	}
    
    }
    
    close(ARCH);
    
    $_ = $t;
}	


#----------------------------------------------------------------------------
# Additions for version 4.0
#----------------------------------------------------------------------------

sub ExtractLastPRFrame
{
    my $t;
    
    print_log("Extracting last PR frame for growing ligand\n");
    
    $t = GetRestartTime($PR_TRR_FILE);
    
    Execute("echo 0 | trjconv -f $PR_TRR_FILE -s $PR_TPR_FILE -o $GENBOX_GRO_FILE -b $t -e $t");
    
}

sub ModifyLigandTopology
{
    my $new_itp_file;
    my $prev_itp_file;
    
    $prev_itp_file = $_[0];
    $new_itp_file = $_[1];
    
    print_log("Changing ligand topology $_[0] by $_[1] in $TOPOLOGY_FILE\n");
    
    $prev_itp_file =~ s/\./\\\./g;
    $prev_itp_file =~ s/\//\\\//g;
    $new_itp_file =~ s/\./\\\./g;
    $new_itp_file =~ s/\//\\\//g;
    Execute("sed -e 's/\\(#include \"\\)\\(\.$prev_itp_file\\)\\(\"\\)/\\1\.$new_itp_file\\3/g' $TOPOLOGY_FILE >top.tmp");
    system("cp top.tmp $TOPOLOGY_FILE");

}

sub GetFileName
{
    my $drive;
    my $dir;
    my $filename;
    my $ext;
    
    ($drive, $dir, $filename) = File::Spec->splitpath(@_[0]);
    
    ($filename, $ext) = split(/\./,$filename);
    
    $_ = $filename;
}